@@ -1,4 +1,4 @@
-;;; -*- mode: Emacs-Lisp; coding: iso-2022-7bit; byte-compile-dynamic: t; -*-
+;;; -*- mode: Emacs-Lisp; coding: iso-2022-7bit; -*-
 ;;; tramp-cache.el --- file information caching for Tramp
 
 ;; Copyright (C) 2000, 2005, 2006 by Free Software Foundation, Inc.
@@ -53,13 +53,16 @@
 
 ;; Pacify byte-compiler.
 (eval-when-compile
+  (require 'cl)
   (autoload 'tramp-message "tramp")
   (autoload 'tramp-tramp-file-p "tramp")
+  ;; We cannot autoload macro `with-parsed-tramp-file-name', it
+  ;; results in problems of byte-compiled code.
+  (autoload 'tramp-dissect-file-name "tramp")
   (autoload 'tramp-file-name-method "tramp")
   (autoload 'tramp-file-name-user "tramp")
   (autoload 'tramp-file-name-host "tramp")
   (autoload 'tramp-file-name-localname "tramp")
-  (autoload 'with-parsed-tramp-file-name "tramp")
   (autoload 'time-stamp-string "time-stamp"))
 
 ;;; -- Cache --
@@ -83,6 +86,7 @@
   "Get the PROPERTY of FILE from the cache context of VEC.
 Returns DEFAULT if not set."
   ;; Unify localname.
+  (setq vec (copy-sequence vec))
   (aset vec 4 (directory-file-name file))
   (let* ((hash (or (gethash vec tramp-cache-data)
 		   (puthash vec (make-hash-table :test 'equal)
@@ -97,6 +101,7 @@ Returns DEFAULT if not set."
   "Set the PROPERTY of FILE to VALUE, in the cache context of VEC.
 Returns VALUE."
   ;; Unify localname.
+  (setq vec (copy-sequence vec))
   (aset vec 4 (directory-file-name file))
   (let ((hash (or (gethash vec tramp-cache-data)
 		  (puthash vec (make-hash-table :test 'equal)
@@ -108,6 +113,7 @@ Returns VALUE."
 (defun tramp-flush-file-property (vec file)
   "Remove all properties of FILE in the cache context of VEC."
   ;; Unify localname.
+  (setq vec (copy-sequence vec))
   (aset vec 4 (directory-file-name file))
   (tramp-message vec 8 "%s" file)
   (remhash vec tramp-cache-data))
@@ -119,7 +125,8 @@ Remove also properties of all files in subdirectories."
   (tramp-message vec 8 "%s" directory)
     (maphash
      '(lambda (key value)
-	(when (string-match directory (tramp-file-name-localname key))
+	(when (and (stringp key)
+		   (string-match directory (tramp-file-name-localname key)))
 	  (remhash key tramp-cache-data)))
      tramp-cache-data)))
 
@@ -164,11 +171,10 @@ FILE must be a local file name on a connection identified via VEC."
 ;; have been changed outside Tramp.
 (defun tramp-cache-before-revert-function ()
   "Flush all Tramp cache properties from buffer-file-name."
-  (let ((bfn (buffer-file-name))
-	;; Pacify byte-compiler.
-	v localname)
+  (let ((bfn (buffer-file-name)))
     (when (and (stringp bfn) (tramp-tramp-file-p bfn))
-      (with-parsed-tramp-file-name bfn nil
+      (let* ((v (tramp-dissect-file-name bfn))
+	     (localname (tramp-file-name-localname v)))
 	(tramp-flush-file-property v localname)))))
 
 (add-hook 'before-revert-hook 'tramp-cache-before-revert-function)
@@ -183,8 +189,11 @@ FILE must be a local file name on a connection identified via VEC."
   "Get the named PROPERTY for the connection.
 KEY identifies the connection, it is either a process or a vector.
 If the value is not set for the connection, returns DEFAULT."
-  ;; Unify key by removing localname from vector.
-  (when (vectorp key) (aset key 4 nil))
+  ;; Unify key by removing localname from vector.  Work with a copy in
+  ;; order to avoid side effects.
+  (when (vectorp key)
+    (setq key (copy-sequence key))
+    (aset key 4 nil))
   (let* ((hash (gethash key tramp-cache-data))
 	 (value (if (hash-table-p hash)
 		   (gethash property hash default)
@@ -196,8 +205,11 @@ If the value is not set for the connection, returns DEFAULT."
   "Set the named PROPERTY of a connection to VALUE.
 KEY identifies the connection, it is either a process or a vector.
 PROPERTY is set persistent when KEY is a vector."
-  ;; Unify key by removing localname from vector.
-  (when (vectorp key) (aset key 4 nil))
+  ;; Unify key by removing localname from vector.  Work with a copy in
+  ;; order to avoid side effects.
+  (when (vectorp key)
+    (setq key (copy-sequence key))
+    (aset key 4 nil))
   (let ((hash (or (gethash key tramp-cache-data)
 		  (puthash key (make-hash-table :test 'equal)
 			    tramp-cache-data))))
@@ -215,8 +227,11 @@ PROPERTY is set persistent when KEY is a vector."
 KEY identifies the connection, it is either a process or a
 vector.  EVENT is not used, it is just applied because this
 function is intended to run also as process sentinel."
-  ;; Unify key by removing localname from vector.
-  (when (vectorp key) (aset key 4 nil))
+  ;; Unify key by removing localname from vector.  Work with a copy in
+  ;; order to avoid side effects.
+  (when (vectorp key)
+    (setq key (copy-sequence key))
+    (aset key 4 nil))
 ;  (tramp-message key 7 "%s" event)
   (remhash key tramp-cache-data))
 
