@@ -1,6 +1,7 @@
 ;;; tramp-smb.el --- Tramp access functions for SMB servers -*- coding: iso-8859-1; -*-
 
-;; Copyright (C) 2002, 2003, 2004, 2005, 2006 Free Software Foundation, Inc.
+;; Copyright (C) 2002, 2003, 2004, 2005, 2006,
+;;   2007 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <michael.albinus@gmx.de>
 ;; Keywords: comm, processes
@@ -75,6 +76,7 @@
   "Regexp used as prompt in smbclient.")
 
 (defconst tramp-smb-errors
+  ;; `regexp-opt' not possible because of first string.
   (mapconcat
    'identity
    '(;; Connection error
@@ -88,24 +90,48 @@
      "ERRnoaccess"
      "ERRnomem"
      "ERRnosuchshare"
-     ;; Windows NT 4.0, Windows 5.0 (Windows 2000), Windows 5.1 (Windows XP)
+     ;; Windows 4.0 (Windows NT), Windows 5.0 (Windows 2000),
+     ;; Windows 5.1 (Windows XP), Windows 5.2 (Windows Server 2003)
      "NT_STATUS_ACCESS_DENIED"
      "NT_STATUS_ACCOUNT_LOCKED_OUT"
      "NT_STATUS_BAD_NETWORK_NAME"
      "NT_STATUS_CANNOT_DELETE"
+     "NT_STATUS_DIRECTORY_NOT_EMPTY"
+     "NT_STATUS_DUPLICATE_NAME"
+     "NT_STATUS_FILE_IS_A_DIRECTORY"
      "NT_STATUS_LOGON_FAILURE"
      "NT_STATUS_NETWORK_ACCESS_DENIED"
      "NT_STATUS_NO_SUCH_FILE"
+     "NT_STATUS_OBJECT_NAME_COLLISION"
      "NT_STATUS_OBJECT_NAME_INVALID"
      "NT_STATUS_OBJECT_NAME_NOT_FOUND"
      "NT_STATUS_SHARING_VIOLATION"
+     "NT_STATUS_TRUSTED_RELATIONSHIP_FAILURE"
      "NT_STATUS_WRONG_PASSWORD")
    "\\|")
   "Regexp for possible error strings of SMB servers.
 Used instead of analyzing error codes of commands.")
 
-(defvar tramp-smb-inodes nil
-  "Keeps virtual inodes numbers for SMB files.")
+(defconst tramp-smb-actions-with-share
+  '((tramp-smb-prompt tramp-action-succeed)
+    (tramp-password-prompt-regexp tramp-action-password)
+    (tramp-wrong-passwd-regexp tramp-action-permission-denied)
+    (tramp-smb-errors tramp-action-permission-denied)
+    (tramp-process-alive-regexp tramp-action-process-alive))
+  "List of pattern/action pairs.
+This list is used for login to SMB servers.
+
+See `tramp-actions-before-shell' for more info.")
+
+(defconst tramp-smb-actions-without-share
+  '((tramp-password-prompt-regexp tramp-action-password)
+    (tramp-wrong-passwd-regexp tramp-action-permission-denied)
+    (tramp-smb-errors tramp-action-permission-denied)
+    (tramp-process-alive-regexp tramp-action-out-of-band))
+  "List of pattern/action pairs.
+This list is used for login to SMB servers.
+
+See `tramp-actions-before-shell' for more info.")
 
 ;; New handlers should be added here.
 (defconst tramp-smb-file-name-handler-alist
@@ -120,8 +146,8 @@ Used instead of analyzing error codes of commands.")
     (directory-file-name . tramp-handle-directory-file-name)
     (directory-files . tramp-smb-handle-directory-files)
     (directory-files-and-attributes . tramp-smb-handle-directory-files-and-attributes)
-    (dired-call-process . tramp-smb-not-handled)
-    (dired-compress-file . tramp-smb-not-handled)
+    (dired-call-process . ignore)
+    (dired-compress-file . ignore)
     ;; `dired-uncache' performed by default handler
     ;; `expand-file-name' not necessary because we cannot expand "~/"
     (file-accessible-directory-p . tramp-smb-handle-file-directory-p)
@@ -139,10 +165,10 @@ Used instead of analyzing error codes of commands.")
     (file-name-nondirectory . tramp-handle-file-name-nondirectory)
     ;; `file-name-sans-versions' performed by default handler
     (file-newer-than-file-p . tramp-smb-handle-file-newer-than-file-p)
-    (file-ownership-preserved-p . tramp-smb-not-handled)
+    (file-ownership-preserved-p . ignore)
     (file-readable-p . tramp-smb-handle-file-exists-p)
     (file-regular-p . tramp-handle-file-regular-p)
-    (file-symlink-p . tramp-smb-not-handled)
+    (file-symlink-p . tramp-handle-file-symlink-p)
     ;; `file-truename' performed by default handler
     (file-writable-p . tramp-smb-handle-file-writable-p)
     (find-backup-file-name . tramp-handle-find-backup-file-name)
@@ -153,15 +179,15 @@ Used instead of analyzing error codes of commands.")
     (load . tramp-handle-load)
     (make-directory . tramp-smb-handle-make-directory)
     (make-directory-internal . tramp-smb-handle-make-directory-internal)
-    (make-symbolic-link . tramp-smb-not-handled)
+    (make-symbolic-link . ignore)
     (rename-file . tramp-smb-handle-rename-file)
-    (set-file-modes . tramp-smb-not-handled)
-    (set-visited-file-modtime . tramp-smb-not-handled)
-    (shell-command . tramp-smb-not-handled)
+    (set-file-modes . ignore)
+    (set-visited-file-modtime . ignore)
+    (shell-command . ignore)
     (substitute-in-file-name . tramp-smb-handle-substitute-in-file-name)
     (unhandled-file-name-directory . tramp-handle-unhandled-file-name-directory)
-    (vc-registered . tramp-smb-not-handled)
-    (verify-visited-file-modtime . tramp-smb-not-handled)
+    (vc-registered . ignore)
+    (verify-visited-file-modtime . ignore)
     (write-region . tramp-smb-handle-write-region)
 )
   "Alist of handler functions for Tramp SMB method.
@@ -178,9 +204,7 @@ First arg specifies the OPERATION, second arg is a list of arguments to
 pass to the OPERATION."
   (let ((fn (assoc operation tramp-smb-file-name-handler-alist)))
     (if fn
-	(if (eq (cdr fn) 'tramp-smb-not-handled)
-	    (apply (cdr fn) operation args)
-	  (save-match-data (apply (cdr fn) args)))
+	(save-match-data (apply (cdr fn) args))
       (tramp-run-real-handler operation args))))
 
 (add-to-list 'tramp-foreign-file-name-handler-alist
@@ -189,14 +213,6 @@ pass to the OPERATION."
 
 ;; File name primitives
 
-(defun tramp-smb-not-handled (operation &rest args)
-  "Default handler for all functions which are disrecarded."
-  (when (stringp (car args))
-    (with-parsed-tramp-file-name (car args) nil
-      (with-file-property v localname operation
-	(tramp-message v 2 "Won't be handled: %s %s" operation args)
-	nil))))
-
 (defun tramp-smb-handle-copy-file
   (filename newname &optional ok-if-already-exists keep-date)
   "Like `copy-file' for tramp files.
@@ -220,13 +236,15 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 		   (file-exists-p newname))
 	  (tramp-error v 'file-already-exists newname))
 
+	;; We must also flush the cache of the directory, because
+	;; file-attributes reads the values from there.
+	(tramp-flush-file-property v (file-name-directory localname))
 	(tramp-flush-file-property v localname)
 	(let ((share (tramp-smb-get-share localname))
 	      (file (tramp-smb-get-localname localname t)))
 	  (unless share
 	    (tramp-error
 	     v 'file-error "Target `%s' must contain a share name" newname))
-	  (tramp-smb-maybe-open-connection v share)
 	  (tramp-message v 0 "Copying file %s to file %s..." filename newname)
 	  (if (tramp-smb-send-command
 	       v (format "put %s \"%s\"" filename file))
@@ -239,55 +257,66 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
   (setq directory (directory-file-name (expand-file-name directory)))
   (when (file-exists-p directory)
     (with-parsed-tramp-file-name directory nil
+      ;; We must also flush the cache of the directory, because
+      ;; file-attributes reads the values from there.
+      (tramp-flush-file-property v (file-name-directory localname))
       (tramp-flush-directory-property v localname)
-      (let ((share (tramp-smb-get-share localname))
-	    (dir (tramp-smb-get-localname (file-name-directory localname) t))
+      (let ((dir (tramp-smb-get-localname (file-name-directory localname) t))
 	    (file (file-name-nondirectory localname)))
-	(tramp-smb-maybe-open-connection v share)
-	(if (and
-	     (tramp-smb-send-command v (format "cd \"%s\"" dir))
-	     (tramp-smb-send-command v (format "rmdir \"%s\"" file)))
-	    ;; Go Home
-	    (tramp-smb-send-command v (format "cd \\"))
-	  ;; Error
-	  (tramp-smb-send-command v (format "cd \\"))
-	  (tramp-error
-	   v 'file-error "Cannot delete directory `%s'" directory))))))
+	(unwind-protect
+	    (unless (and
+		     (tramp-smb-send-command v (format "cd \"%s\"" dir))
+		     (tramp-smb-send-command v (format "rmdir \"%s\"" file)))
+	      ;; Error
+	      (with-current-buffer (tramp-get-connection-buffer v)
+		(goto-char (point-min))
+		(search-forward-regexp tramp-smb-errors nil t)
+		(tramp-error
+		 v 'file-error "%s `%s'" (match-string 0) directory)))
+	  ;; Always go home
+	  (tramp-smb-send-command v (format "cd \\")))))))
 
 (defun tramp-smb-handle-delete-file (filename)
   "Like `delete-file' for tramp files."
   (setq filename (expand-file-name filename))
   (when (file-exists-p filename)
     (with-parsed-tramp-file-name filename nil
+      ;; We must also flush the cache of the directory, because
+      ;; file-attributes reads the values from there.
+      (tramp-flush-file-property v (file-name-directory localname))
       (tramp-flush-file-property v localname)
-      (let ((share (tramp-smb-get-share localname))
-	    (dir (tramp-smb-get-localname (file-name-directory localname) t))
+      (let ((dir (tramp-smb-get-localname (file-name-directory localname) t))
 	    (file (file-name-nondirectory localname)))
-	(tramp-smb-maybe-open-connection v share)
-	(if (and
-	     (tramp-smb-send-command v (format "cd \"%s\"" dir))
-	     (tramp-smb-send-command v (format "rm \"%s\"" file)))
-	    ;; Go Home
-	    (tramp-smb-send-command v (format "cd \\"))
-	  ;; Error
-	  (tramp-smb-send-command v (format "cd \\"))
-	  (tramp-error v 'file-error "Cannot delete file `%s'" filename))))))
+	(unwind-protect
+	    (unless (and
+		     (tramp-smb-send-command v (format "cd \"%s\"" dir))
+		     (tramp-smb-send-command v (format "rm \"%s\"" file)))
+	      ;; Error
+	      (with-current-buffer (tramp-get-connection-buffer v)
+		(goto-char (point-min))
+		(search-forward-regexp tramp-smb-errors nil t)
+		(tramp-error
+		 v 'file-error "%s `%s'" (match-string 0) filename)))
+	  ;; Always go home
+	  (tramp-smb-send-command v (format "cd \\")))))))
 
 (defun tramp-smb-handle-directory-files
   (directory &optional full match nosort)
   "Like `directory-files' for tramp files."
-  (let ((result
-	 (mapcar
-	  (lambda (x)
-	    (directory-file-name
-	     (if full (concat (file-name-as-directory directory) x) x)))
-	  (file-name-all-completions "" directory))))
+  (let ((result (mapcar 'directory-file-name
+			(file-name-all-completions "" directory))))
     ;; Discriminate with regexp
     (when match
       (setq result
 	    (delete nil
 		    (mapcar (lambda (x) (when (string-match match x) x))
 			    result))))
+    ;; Append directory
+    (when full
+      (setq result
+	    (mapcar
+	     (lambda (x) (concat (file-name-as-directory directory) x))
+	     result)))
     ;; Sort them if necessary
     (unless nosort (setq result (sort result 'string-lessp)))
     ;; That's it
@@ -306,16 +335,20 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 
 (defun tramp-smb-handle-file-attributes (filename &optional id-format)
   "Like `file-attributes' for tramp files."
+  ;; Reading just the filename entry via "dir localname" is not
+  ;; possible, because when filename is a directory, some smbclient
+  ;; versions return the content of the directory, and other versions
+  ;; don't.  Therefore, the whole content of the upper directory is
+  ;; retrieved, and the entry of the filename is extracted from.
   (with-parsed-tramp-file-name filename nil
     (with-file-property v localname (format "file-attributes-%s" id-format)
-      (let* ((share (tramp-smb-get-share localname))
-	     (file (tramp-smb-get-localname localname nil))
-	     (entries (tramp-smb-get-file-entries v share file))
+      (let* ((entries (tramp-smb-get-file-entries
+		       (file-name-directory filename)))
 	     (entry (and entries
-			 (assoc (file-name-nondirectory file) entries)))
+			 (assoc (file-name-nondirectory filename) entries)))
 	     (uid (if (and id-format (equal id-format 'string)) "nobody" -1))
 	     (gid (if (and id-format (equal id-format 'string)) "nogroup" -1))
-	     (inode (tramp-smb-get-inode share file))
+	     (inode (tramp-get-inode filename))
 	     (device (tramp-get-device v)))
 
         ;; Check result.
@@ -336,40 +369,23 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
 
 (defun tramp-smb-handle-file-directory-p (filename)
   "Like `file-directory-p' for tramp files."
-  (with-parsed-tramp-file-name filename nil
-    (with-file-property v localname "file-directory-p"
-      (let* ((share (tramp-smb-get-share localname))
-	     (file (tramp-smb-get-localname localname nil))
-	     (entries (tramp-smb-get-file-entries v share file))
-	     (entry (and entries
-			 (assoc (file-name-nondirectory file) entries))))
-	(and entry
-	     (string-match "d" (nth 1 entry))
-	     t)))))
+  (and (file-exists-p filename)
+       (eq ?d (aref (nth 8 (file-attributes filename)) 0))))
 
 (defun tramp-smb-handle-file-exists-p (filename)
   "Like `file-exists-p' for tramp files."
-  (with-parsed-tramp-file-name filename nil
-    (with-file-property v localname "file-exists-p"
-      (let* ((share (tramp-smb-get-share localname))
-	     (file (tramp-smb-get-localname localname nil))
-	     (entries (tramp-smb-get-file-entries v share file)))
-	(and entries
-	     (member (file-name-nondirectory file) (mapcar 'car entries))
-	     t)))))
+  (not (null (file-attributes filename))))
 
 (defun tramp-smb-handle-file-local-copy (filename)
   "Like `file-local-copy' for tramp files."
   (with-parsed-tramp-file-name filename nil
-    (let ((share (tramp-smb-get-share localname))
-	  (file (tramp-smb-get-localname localname t))
+    (let ((file (tramp-smb-get-localname localname t))
 	  (tmpfil (tramp-make-temp-file)))
       (unless (file-exists-p filename)
 	(tramp-error
 	 v 'file-error
 	 "Cannot make local copy of non-existing file `%s'" filename))
       (tramp-message v 4 "Fetching %s to tmp file %s..." filename tmpfil)
-      (tramp-smb-maybe-open-connection v share)
       (if (tramp-smb-send-command v (format "get \"%s\" %s" file tmpfil))
 	  (tramp-message
 	   v 4 "Fetching %s to tmp file %s...done" filename tmpfil)
@@ -387,10 +403,7 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
    (with-parsed-tramp-file-name directory nil
      (with-file-property v localname "file-name-all-completions"
        (save-match-data
-	 (let* ((share (tramp-smb-get-share localname))
-		(file (file-name-as-directory
-		       (tramp-smb-get-localname localname nil)))
-		(entries (tramp-smb-get-file-entries v share file)))
+	 (let ((entries (tramp-smb-get-file-entries directory)))
 	   (mapcar
 	    (lambda (x)
 	      (list
@@ -404,48 +417,47 @@ KEEP-DATE is not handled in case NEWNAME resides on an SMB server."
   (cond
    ((not (file-exists-p file1)) nil)
    ((not (file-exists-p file2)) t)
-   (t (tramp-smb-time-less-p (file-attributes file2)
-			     (file-attributes file1)))))
+   (t (tramp-time-less-p (nth 5 (file-attributes file2))
+			 (nth 5 (file-attributes file1))))))
 
 (defun tramp-smb-handle-file-writable-p (filename)
   "Like `file-writable-p' for tramp files."
-  (if (not (file-exists-p filename))
-      (let ((dir (file-name-directory filename)))
-	(and (file-exists-p dir)
-	     (file-writable-p dir)))
-    (with-parsed-tramp-file-name filename nil
-      (with-file-property v localname "file-writable-p"
-	(let* ((share (tramp-smb-get-share localname))
-	       (file (tramp-smb-get-localname localname nil))
-	       (entries (tramp-smb-get-file-entries v share file))
-	       (entry (and entries
-			   (assoc (file-name-nondirectory file) entries))))
-	  (and share entry
-	       (string-match "w" (nth 1 entry))
-	       t))))))
+  (if (file-exists-p filename)
+      (string-match "w" (or (nth 8 (file-attributes filename)) ""))
+    (let ((dir (file-name-directory filename)))
+      (and (file-exists-p dir)
+	   (file-writable-p dir)))))
 
 (defun tramp-smb-handle-insert-directory
   (filename switches &optional wildcard full-directory-p)
-  "Like `insert-directory' for tramp files.
-WILDCARD and FULL-DIRECTORY-P are not handled."
+  "Like `insert-directory' for tramp files."
   (setq filename (expand-file-name filename))
-  (when (file-directory-p filename)
-    ;; This check is a little bit strange, but in `dired-add-entry'
-    ;; this function is called with a non-directory ...
+  (when full-directory-p
+    ;; Called from `dired-add-entry'.
     (setq filename (file-name-as-directory filename)))
   (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-property v localname)
+    (tramp-flush-file-property v (file-name-directory localname))
     (save-match-data
-      (let* ((share (tramp-smb-get-share localname))
-	     (file (tramp-smb-get-localname localname nil))
-	     (entries (tramp-smb-get-file-entries v share file)))
+      (let ((base (file-name-nondirectory filename))
+	    ;; We should not destroy the cache entry.
+	    (entries (copy-sequence
+		      (tramp-smb-get-file-entries
+		       (file-name-directory filename)))))
 
-	;; Delete dummy "" entry, useless entries
+	;; Filter entries.
 	(setq entries
-	      (if (file-directory-p filename)
-		  (delq (assoc "" entries) entries)
-		;; We just need the only and only entry FILENAME.
-		(list (assoc (file-name-nondirectory filename) entries))))
+	      (delq
+	       nil
+	       (if (or wildcard (zerop (length base)))
+		   ;; Check for matching entries.
+		   (mapcar
+		    (lambda (x)
+		      (when (string-match
+			     (format "^%s" (regexp-quote base)) (nth 0 x))
+			x))
+		    entries)
+		 ;; We just need the only and only entry FILENAME.
+		 (list (assoc base entries)))))
 
 	;; Sort entries
 	(setq entries
@@ -454,30 +466,31 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
 	       (lambda (x y)
 		 (if (string-match "t" switches)
 		     ;; Sort by date.
-		     (tramp-smb-time-less-p (nth 3 y) (nth 3 x))
+		     (tramp-time-less-p (nth 3 y) (nth 3 x))
 		   ;; Sort by name.
 		   (string-lessp (nth 0 x) (nth 0 y))))))
 
 	;; Print entries.
 	(mapcar
 	 (lambda (x)
-	   (insert
-	    (format
-	     "%10s %3d %-8s %-8s %8s %s %s\n"
-	     (nth 1 x) ; mode
-	     1 "nobody" "nogroup"
-	     (nth 2 x) ; size
-	     (format-time-string
-	      (if (tramp-smb-time-less-p
-		   (tramp-smb-time-subtract (current-time) (nth 3 x))
-		   tramp-smb-half-a-year)
-		  "%b %e %R"
-		"%b %e  %Y")
-	      (nth 3 x)) ; date
-	     (nth 0 x))) ; file name
-	   (forward-line)
-	   (beginning-of-line))
-	 entries)))))
+	   (when (not (zerop (length (nth 0 x))))
+	     (insert
+	      (format
+	       "%10s %3d %-8s %-8s %8s %s %s\n"
+	       (nth 1 x) ; mode
+	       1 "nobody" "nogroup"
+	       (nth 2 x) ; size
+	       (format-time-string
+		(if (tramp-time-less-p
+		     (tramp-time-subtract (current-time) (nth 3 x))
+		     tramp-half-a-year)
+		    "%b %e %R"
+		  "%b %e  %Y")
+		(nth 3 x)) ; date
+	       (nth 0 x))) ; file name
+	     (forward-line)
+	     (beginning-of-line)))
+	   entries)))))
 
 (defun tramp-smb-handle-make-directory (dir &optional parents)
   "Like `make-directory' for tramp files."
@@ -504,11 +517,12 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
     (setq directory (concat default-directory directory)))
   (with-parsed-tramp-file-name directory nil
     (save-match-data
-      (let* ((share (tramp-smb-get-share localname))
-	     (file (tramp-smb-get-localname localname nil)))
+      (let* ((file (tramp-smb-get-localname localname t)))
 	(when (file-directory-p (file-name-directory directory))
-	  (tramp-smb-maybe-open-connection v share)
-	  (tramp-smb-send-command v (format "mkdir \"%s\"" file)))
+	  (tramp-smb-send-command v (format "mkdir \"%s\"" file))
+	  ;; We must also flush the cache of the directory, because
+	  ;; file-attributes reads the values from there.
+	  (tramp-flush-file-property v (file-name-directory localname)))
 	(unless (file-directory-p directory)
 	  (tramp-error
 	   v 'file-error "Couldn't make directory %s" directory))))))
@@ -534,10 +548,11 @@ WILDCARD and FULL-DIRECTORY-P are not handled."
 	(when (and (not ok-if-already-exists)
 		   (file-exists-p newname))
 	  (tramp-error v 'file-already-exists newname))
+	;; We must also flush the cache of the directory, because
+	;; file-attributes reads the values from there.
+	(tramp-flush-file-property v (file-name-directory localname))
 	(tramp-flush-file-property v localname)
-	(let ((share (tramp-smb-get-share localname))
-	      (file (tramp-smb-get-localname localname t)))
-	  (tramp-smb-maybe-open-connection v share)
+	(let ((file (tramp-smb-get-localname localname t)))
 	  (tramp-message v 0 "Copying file %s to file %s..." filename newname)
 	  (if (tramp-smb-send-command v (format "put %s \"%s\"" filename file))
 	      (tramp-message
@@ -567,9 +582,11 @@ Catches errors for shares like \"C$/\", which are common in Microsoft Windows."
       (unless (y-or-n-p (format "File %s exists; overwrite anyway? "
 				filename))
 	(tramp-error v 'file-error "File not overwritten")))
+    ;; We must also flush the cache of the directory, because
+    ;; file-attributes reads the values from there.
+    (tramp-flush-file-property v (file-name-directory localname))
     (tramp-flush-file-property v localname)
-    (let ((share (tramp-smb-get-share localname))
-	  (file (tramp-smb-get-localname localname t))
+    (let ((file (tramp-smb-get-localname localname t))
 	  (curbuf (current-buffer))
 	  tmpfil)
       ;; Write region into a tmp file.
@@ -583,7 +600,6 @@ Catches errors for shares like \"C$/\", which are common in Microsoft Windows."
 	   (list start end tmpfil append 'no-message lockname confirm)
 	 (list start end tmpfil append 'no-message lockname)))
 
-      (tramp-smb-maybe-open-connection v share)
       (tramp-message v 5 "Writing tmp file %s to file %s..." tmpfil filename)
       (if (tramp-smb-send-command v (format "put %s \"%s\"" tmpfil file))
 	  (tramp-message
@@ -632,51 +648,53 @@ If CONVERT is non-nil exchange \"/\" by \"\\\\\"."
 
 ;; Share names of a host are cached. It is very unlikely that the
 ;; shares do change during connection.
-(defun tramp-smb-get-file-entries (vec share localname)
-  "Read entries which match LOCALNAME.
+(defun tramp-smb-get-file-entries (directory)
+  "Read entries which match DIRECTORY.
 Either the shares are listed, or the `dir' command is executed.
-Only entries matching the localname are returned.
 Result is a list of (LOCALNAME MODE SIZE MONTH DAY TIME YEAR)."
-  (with-current-buffer (tramp-get-buffer vec)
-    (let ((base (or (and (> (length localname) 0)
-			 (string-match "\\([^/]+\\)$" localname)
-			 (regexp-quote (match-string 1 localname)))
-		    ""))
-	  (cache (tramp-get-connection-property vec "share-cache" nil))
-	  res entry)
-      (if (and (not share) cache)
-	  ;; Return cached shares
-	  (setq res cache)
-	;; Read entries
-	(tramp-smb-maybe-open-connection vec share)
-	(when share
-	  (tramp-smb-send-command
-	   vec
-	   (format
-	    "dir %s"
-	    (if (zerop (length localname)) "" (concat "\"" localname "*\"")))))
-	(goto-char (point-min))
-	;; Loop the listing
-	(unless (re-search-forward tramp-smb-errors nil t)
-	  (while (not (eobp))
-	    (setq entry (tramp-smb-read-file-entry share))
-	    (forward-line)
-	    (when entry (add-to-list 'res entry))))
-	(unless share
-	  ;; Cache share entries
-	  (tramp-set-connection-property vec "share-cache" res)))
+  (with-parsed-tramp-file-name directory nil
+    (setq localname (or localname "/"))
+    (with-file-property v localname "file-entries"
+      (with-current-buffer (tramp-get-buffer v)
+	(let* ((share (tramp-smb-get-share localname))
+	       (file (tramp-smb-get-localname localname nil))
+	       (cache (tramp-get-connection-property v "share-cache" nil))
+	       res entry)
 
-      ;; Add directory itself
-      (add-to-list 'res '("" "drwxrwxrwx" 0 (0 0)))
+	  (if (and (not share) cache)
+	      ;; Return cached shares
+	      (setq res cache)
 
-      ;; There's a very strange error (debugged with XEmacs 21.4.14)
-      ;; If there's no short delay, it returns nil.  No idea about
-      (when (featurep 'xemacs) (sleep-for 0.01))
+	    ;; Read entries
+	    (setq file (file-name-as-directory file))
+	    (when (string-match "^\\./" file)
+	      (setq file (substring file 1)))
+	    (if share
+		(tramp-smb-send-command v (format "dir \"%s*\"" file))
+	      ;; `tramp-smb-maybe-open-connection' lists also the share names
+	      (tramp-smb-maybe-open-connection v))
 
-      ;; Check for matching entries
-      (delq nil (mapcar
-		 (lambda (x) (and (string-match base (nth 0 x)) x))
-		 res)))))
+	    ;; Loop the listing
+	    (goto-char (point-min))
+	    (unless (re-search-forward tramp-smb-errors nil t)
+	      (while (not (eobp))
+		(setq entry (tramp-smb-read-file-entry share))
+		(forward-line)
+		(when entry (add-to-list 'res entry))))
+
+	    ;; Cache share entries
+	    (unless share
+	      (tramp-set-connection-property v "share-cache" res)))
+
+	  ;; Add directory itself
+	  (add-to-list 'res '("" "drwxrwxrwx" 0 (0 0)))
+
+	  ;; There's a very strange error (debugged with XEmacs 21.4.14)
+	  ;; If there's no short delay, it returns nil.  No idea about.
+	  (when (featurep 'xemacs) (sleep-for 0.01))
+
+	  ;; Return entries
+	  (delq nil res))))))
 
 ;; Return either a share name (if SHARE is nil), or a file name
 ;;
@@ -744,7 +762,7 @@ Result is the list (LOCALNAME MODE SIZE MTIME)."
     (if (not share)
 
 	;; Read share entries.
-	(when (string-match "^\\s-+\\(\\S-+\\)\\s-+Disk" line)
+	(when (string-match "^\\s-+\\(\\S-\\(.*\\S-\\)?\\)\\s-+Disk" line)
 	  (setq localname (match-string 1 line)
 		mode "dr-xr-xr-x"
 		size 0))
@@ -815,126 +833,112 @@ Result is the list (LOCALNAME MODE SIZE MTIME)."
 	    (if (and sec min hour day month year)
 		(encode-time
 		 sec min hour day
-		 (cdr (assoc (downcase month) tramp-smb-parse-time-months))
+		 (cdr (assoc (downcase month) tramp-parse-time-months))
 		 year)
 	      '(0 0)))
       (list localname mode size mtime))))
 
-;; Inodes don't exist for SMB files.  Therefore we must generate virtual ones.
-;; Used in `find-buffer-visiting'.
-;; The method applied might be not so efficient (Ange-FTP uses hashes). But
-;; performance isn't the major issue given that file transfer will take time.
-
-(defun tramp-smb-get-inode (share file)
-  "Returns the virtual inode number.
-If it doesn't exist, generate a new one."
-  (let ((string (concat share "/" (directory-file-name file))))
-    (unless (assoc string tramp-smb-inodes)
-      (add-to-list 'tramp-smb-inodes
-		   (list string (length tramp-smb-inodes))))
-    (nth 1 (assoc string tramp-smb-inodes))))
-
 
 ;; Connection functions
 
 (defun tramp-smb-send-command (vec command)
   "Send the COMMAND to connection VEC.
 Returns nil if there has been an error message from smbclient."
+  (tramp-smb-maybe-open-connection vec)
   (tramp-message vec 6 "%s" command)
   (tramp-send-string vec command)
   (tramp-smb-wait-for-output vec))
 
-(defun tramp-smb-maybe-open-connection (vec share)
-  "Maybe open a connection to HOST, logging in as USER, using `tramp-smb-program'.
+(defun tramp-smb-maybe-open-connection (vec)
+  "Maybe open a connection to HOST, log in as USER, using `tramp-smb-program'.
 Does not do anything if a connection is already open, but re-opens the
 connection if a previous connection has died for some reason."
-  (let ((process-connection-type tramp-process-connection-type)
-	(p (get-buffer-process (tramp-get-buffer vec))))
+  (let* ((share (tramp-smb-get-share (tramp-file-name-localname vec)))
+	 (buf (tramp-get-buffer vec))
+	 (p (get-buffer-process buf)))
     ;; Check whether it is still the same share
     (unless
-	(and p (processp p)
-	     (memq (process-status p) '(run open))
+	(and p (processp p) (memq (process-status p) '(run open))
 	     (string-equal
 	      share
 	      (tramp-get-connection-property p "smb-share" "")))
-      (when (and p (processp p))
-	(delete-process p))
-      (tramp-smb-open-connection vec share))))
 
-(defun tramp-smb-open-connection (vec share)
-  "Open a connection using `tramp-smb-program'.
-This starts the command `smbclient //HOST/SHARE -U USER', then waits
-for a remote password prompt.  It queries the user for the password,
-then sends the password to the remote host.
+      (save-match-data
+	;; There might be unread output from checking for share names.
+	(when buf (with-current-buffer buf (erase-buffer)))
+	(when (and p (processp p)) (delete-process p))
 
-Domain names in USER and port numbers in HOST are acknowledged."
-;; Called from `tramp-smb-maybe-open-connection', which has set the
-;; current buffer.
-  (unless (let ((default-directory
-		  (tramp-temporary-file-directory)))
-	    (executable-find tramp-smb-program))
-    (error "Cannot find command %s in %s" tramp-smb-program exec-path))
+	(unless (let ((default-directory
+			(tramp-temporary-file-directory)))
+		  (executable-find tramp-smb-program))
+	  (error "Cannot find command %s in %s" tramp-smb-program exec-path))
 
-  (save-match-data
-    (let* ((user (tramp-file-name-user vec))
-	   (host (tramp-file-name-host vec))
-	   (real-user user)
-	   (real-host host)
-	   domain port args)
+	(let* ((user (tramp-file-name-user vec))
+	       (host (tramp-file-name-host vec))
+	       (real-user user)
+	       (real-host host)
+	       domain port args)
 
-      ;; Check for domain ("user%domain") and port ("host#port").
-      (when (and user (string-match "\\(.+\\)%\\(.+\\)" user))
-	(setq real-user (or (match-string 1 user) user)
-	      domain (match-string 2 user)))
+	  ;; Check for domain ("user%domain") and port ("host#port").
+	  (when (and user (string-match "\\(.+\\)%\\(.+\\)" user))
+	    (setq real-user (or (match-string 1 user) user)
+		  domain (match-string 2 user)))
 
-      (when (and host (string-match "\\(.+\\)#\\(.+\\)" host))
-	(setq real-host (or (match-string 1 host) host)
-	      port (match-string 2 host)))
+	  (when (and host (string-match "\\(.+\\)#\\(.+\\)" host))
+	    (setq real-host (or (match-string 1 host) host)
+		  port (match-string 2 host)))
 
-      (if share
-	  (setq args (list (concat "//" real-host "/" share)))
-	(setq args (list "-L" real-host )))
+	  (if share
+	      (setq args (list (concat "//" real-host "/" share)))
+	    (setq args (list "-L" real-host )))
 
-      (if real-user
-	  (setq args (append args (list "-U" real-user)))
-	(setq args (append args (list "-N"))))
+	  (if (not (zerop (length real-user)))
+	      (setq args (append args (list "-U" real-user)))
+	    (setq args (append args (list "-N"))))
 
-      (when domain (setq args (append args (list "-W" domain))))
-      (when port   (setq args (append args (list "-p" port))))
+	  (when domain (setq args (append args (list "-W" domain))))
+	  (when port   (setq args (append args (list "-p" port))))
+	  (setq args (append args (list "-s" "/dev/null")))
 
-      ;; OK, let's go.
-      (tramp-message
-       vec 3 "Opening connection for //%s@%s/%s..." user host (or share ""))
+	  ;; OK, let's go.
+	  (tramp-message
+	   vec 3 "Opening connection for //%s%s/%s..."
+	   (if (not (zerop (length user))) (concat user "@") "")
+	   host (or share ""))
 
-      (let* ((coding-system-for-read nil)
-	     (p (let ((default-directory (tramp-temporary-file-directory)))
-		  (apply #'start-process
-			 (tramp-buffer-name vec) (tramp-get-buffer vec)
-			 tramp-smb-program args))))
+	  (let* ((coding-system-for-read nil)
+		 (process-connection-type tramp-process-connection-type)
+		 (p (let ((default-directory (tramp-temporary-file-directory)))
+		      (apply #'start-process
+			     (tramp-buffer-name vec) (tramp-get-buffer vec)
+			     tramp-smb-program args))))
 
-	(tramp-message vec 6 "%s" (mapconcat 'identity (process-command p) " "))
-	(set-process-sentinel p 'tramp-flush-connection-property)
-	(tramp-set-process-query-on-exit-flag p nil)
-	(tramp-set-connection-property p "smb-share" share)
+	    (tramp-message
+	     vec 6 "%s" (mapconcat 'identity (process-command p) " "))
+	    (set-process-sentinel p 'tramp-flush-connection-property)
+	    (tramp-set-process-query-on-exit-flag p nil)
+	    (tramp-set-connection-property p "smb-share" share)
 
-	;; Set variables for computing the prompt for reading password.
-	(setq tramp-current-method tramp-smb-method
-	      tramp-current-user user
-	      tramp-current-host host)
+	    ;; Set variables for computing the prompt for reading password.
+	    (setq tramp-current-method tramp-smb-method
+		  tramp-current-user user
+		  tramp-current-host host)
 
-	;; Set chunksize.  Otherwise, `tramp-send-string' might try it itself.
-	(tramp-set-connection-property p "chunksize" tramp-chunksize)
+	    ;; Set chunksize.  Otherwise, `tramp-send-string' might
+	    ;; try it itself.
+	    (tramp-set-connection-property p "chunksize" tramp-chunksize)
 
-        ;; Send password.
-	(when real-user
-	  (tramp-message vec 3 "Sending password")
-	  (tramp-enter-password p))
+	    ;; Play login scenario.
+	    (tramp-process-actions
+	     p vec
+	     (if share
+		 tramp-smb-actions-with-share
+	       tramp-smb-actions-without-share))
 
-	(unless (tramp-smb-wait-for-output vec)
-	  (tramp-clear-passwd)
-	  (tramp-error
-	   vec 'file-error
-	   "Cannot open connection //%s@%s/%s" user host (or share "")))))))
+	    (tramp-message
+	     vec 3 "Opening connection for //%s%s/%s...done"
+	     (if (not (zerop (length user))) (concat user "@") "")
+	     host (or share ""))))))))
 
 ;; We don't use timeouts.  If needed, the caller shall wrap around.
 (defun tramp-smb-wait-for-output (vec)
@@ -963,52 +967,29 @@ Returns nil if an error message has appeared."
 	(goto-char (point-min))
 	(setq err (re-search-forward tramp-smb-errors nil t)))
 
+      ;; When the process is still alive, read pending output.
+      (while (and (not found) (memq (process-status p) '(run open)))
+
+	;; Accept pending output.
+	(tramp-accept-process-output p)
+
+	;; Search for prompt.
+	(goto-char (point-min))
+	(setq found (re-search-forward tramp-smb-prompt nil t)))
+
       ;; Return value is whether no error message has appeared.
       (tramp-message vec 6 "\n%s" (buffer-string))
       (not err))))
 
 
-;; Snarfed code from time-date.el and parse-time.el
-
-(defconst tramp-smb-half-a-year '(241 17024)
-"Evaluated by \"(days-to-time 183)\".")
-
-(defconst tramp-smb-parse-time-months
-  '(("jan" . 1) ("feb" . 2) ("mar" . 3)
-    ("apr" . 4) ("may" . 5) ("jun" . 6)
-    ("jul" . 7) ("aug" . 8) ("sep" . 9)
-    ("oct" . 10) ("nov" . 11) ("dec" . 12))
-  "Alist mapping month names to integers.")
-
-(defun tramp-smb-time-less-p (t1 t2)
-  "Say whether time value T1 is less than time value T2."
-  (unless t1 (setq t1 '(0 0)))
-  (unless t2 (setq t2 '(0 0)))
-  (or (< (car t1) (car t2))
-      (and (= (car t1) (car t2))
-	   (< (nth 1 t1) (nth 1 t2)))))
-
-(defun tramp-smb-time-subtract (t1 t2)
-  "Subtract two time values.
-Return the difference in the format of a time value."
-  (unless t1 (setq t1 '(0 0)))
-  (unless t2 (setq t2 '(0 0)))
-  (let ((borrow (< (cadr t1) (cadr t2))))
-    (list (- (car t1) (car t2) (if borrow 1 0))
-	  (- (+ (if borrow 65536 0) (cadr t1)) (cadr t2)))))
-
-
 (provide 'tramp-smb)
 
 ;;; TODO:
 
-;; * Provide a local smb.conf. The default one might not be readable.
 ;; * Error handling in case password is wrong.
 ;; * Read password from "~/.netrc".
 ;; * Return more comprehensive file permission string.  Think whether it is
 ;;   possible to implement `set-file-modes'.
-;; * Handle WILDCARD and FULL-DIRECTORY-P in
-;;   `tramp-smb-handle-insert-directory'.
 ;; * Handle links (FILENAME.LNK).
 ;; * Maybe local tmp files should have the same extension like the original
 ;;   files.  Strange behaviour with jka-compr otherwise?
